/* Partycle Popper 
 *  version 1.1, December 16th, 2007
 *
 *  Copyright (C) 2007 Jari KOmppa
 *
 *  This software is provided 'as-is', without any express or implied
 *  warranty.  In no event will the authors be held liable for any damages
 *  arising from the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, subject to the following restrictions:
 *
 *  1. The origin of this software must not be misrepresented; you must not
 *     claim that you wrote the original software. If you use this software
 *     in a product, an acknowledgment in the product documentation would be
 *     appreciated but is not required.
 *  2. Altered source versions must be plainly marked as such, and must not be
 *     misrepresented as being the original software.
 *  3. This notice may not be removed or altered from any source distribution.
 *
 * Jari Komppa http://iki.fi/sol/
 *
 *************************************
 * This is my entry to the Ludum Dare 48h game design competition number 10.
 *
 * Here be dragons. 
 *
 * It's a mess. 
 *
 * I didn't try to clean things up.. probably one evenings' worth work would
 * turn this into a fairly readable source.
 *
 * The code should be relatively portable, the only major non-opensource 
 * external lib is fmod, and you can turn that off by not defining AUDIO_ENABLED.
 *
 * Cheers,
 *    Jari
 */

#ifdef _MSC_VER
#include <windows.h> // needed to get GL stuff to work
#endif
#include <SDL.h>
#include "glee/glee.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "loadtexture.h"
#include "MersenneTwister.h"
#define AUDIO_ENABLED
#ifdef AUDIO_ENABLED
#include "fmod.h"
#endif

#define TITLE "Partycle Popper 1.1"

#pragma warning(disable:4244) 
#pragma warning(disable:4305) 

#define M_PI 3.1415926535897932384626433832795

#ifdef AUDIO_ENABLED
FSOUND_SAMPLE *s_badpop, *s_click, *s_gameover, *s_pop, *s_shoot;
FMUSIC_MODULE *s_midi[5];
#endif

MersenneTwister visualRand, physicsRand;

struct Particle
{
    int live;
    float x, y;
    float dirx, diry;  
    int tex;
};

struct Partycle
{
    int live;
    float x, y;
    int dirmask;
};

struct Bullet
{
    int live;
    float x, y;
    float dirx, diry;
    int value;
    int tex;
};

struct Popup
{
    int live;
    float x, y;
    char * text;
};

enum gamestates
{
    STATE_MENU,
    STATE_INGAME,
    STATE_DYING,
    STATE_STAGECLEAR
};

#define MAX_PARTYCLES 200
#define MAX_BULLETS (MAX_PARTYCLES * 8)
#define MAX_PARTICLES 5000
#define PARTICLES_PER_EXPLOSION 50
#define MAX_POPUPS 50

Partycle partycle[MAX_PARTYCLES];
Particle particle[MAX_PARTICLES];
Bullet bullet[MAX_BULLETS];
Popup popup[MAX_POPUPS];

enum keystates
{
    KEY_UP    = 1,
    KEY_DOWN  = 2,
    KEY_LEFT  = 4,
    KEY_RIGHT = 8,
    KEY_FIRE  = 16
};


int tex_ship, tex_red, tex_base, tex_dir[8], tex_font, tex_bg, tex_green, tex_black;
int gKeyState = 0;
float gPlayerDirection = 0;
float gPlayerDirectionInc = 0;
float gPlayerX = 320;
float gPlayerY = 240;
float gPlayerXi = 0;
float gPlayerYi = 0;
int gScore = 0;
int gStage = 1;
int gGameState = STATE_MENU;
int gStateStartTick = 0;
int last_tick = 0;
int gMouseX, gMouseY;
int screenwidth = 0;
int screenheight = 0;
int mastermusicswitch = 1;

void spawn_popup(char * text, float x, float y)
{
    int i = 0;
    while (i < MAX_POPUPS && popup[i].live) i++;
    if (i >= MAX_POPUPS) return;
    popup[i].live = 100;
    popup[i].text = strdup(text);
    popup[i].y = y;
    popup[i].x = x - strlen(text) * 4;
}

int gPartycleIndex;

void spawn_partycle(float x, float y, int dirmask)
{
    if (gPartycleIndex >= MAX_PARTYCLES)
        return;
    if (x > 630 || x < 10 || y > 470 || y < 10)
        return;
    int basetime = 2000 - gStage * 5;
    if (basetime < 1000) basetime = 1000;
    partycle[gPartycleIndex].live = basetime + gPartycleIndex * 10;
    partycle[gPartycleIndex].dirmask = dirmask;
    // make sure partycle isn't in the spawn point
    while (x > 320-16 && x < 320+16 && y > 240-16 && y < 240 + 16)
    {
        x += (physicsRand.genrand_int31() % 8) - 4;
        y += (physicsRand.genrand_int31() % 8) - 4;
    }
    partycle[gPartycleIndex].x = x;
    partycle[gPartycleIndex].y = y;
    gPartycleIndex++;
}

void spawn_partycle_square(float x, float y, float r, int mode)
{
    if (mode == 0)
    {
        spawn_partycle(x-r, y+r, 4);
        spawn_partycle(x+r, y+r, 1);
        spawn_partycle(x-r, y-r, 16);
        spawn_partycle(x+r, y-r, 64);
    }
    else
    if (mode == 1)
    {
        spawn_partycle(x-r, y+r, 4 | 64);
        spawn_partycle(x+r, y+r, 1 | 16);
        spawn_partycle(x-r, y-r, 16 | 1);
        spawn_partycle(x+r, y-r, 64 | 4);
    }
    else
    {
        int i;
        int extra_arrows;
        extra_arrows = 0; for (i = 0; i < mode; i++) extra_arrows |= (1 << (physicsRand.genrand_int31() & 7));
        spawn_partycle(x-r, y+r, 4 | extra_arrows);
        extra_arrows = 0; for (i = 0; i < mode; i++) extra_arrows |= (1 << (physicsRand.genrand_int31() & 7));
        spawn_partycle(x+r, y+r, 1 | extra_arrows);
        extra_arrows = 0; for (i = 0; i < mode; i++) extra_arrows |= (1 << (physicsRand.genrand_int31() & 7));
        spawn_partycle(x-r, y-r, 16 | extra_arrows);
        extra_arrows = 0; for (i = 0; i < mode; i++) extra_arrows |= (1 << (physicsRand.genrand_int31() & 7));
        spawn_partycle(x+r, y-r, 64 | extra_arrows);
    }
}

int gen_partycle_dirmask(int mode)
{
    int dirmask;
    switch (mode)
    {
    case 0:
        dirmask = 0;
        break;
    case 1:
        if (gPartycleIndex & 1)
            dirmask = 1+16;
        else
            dirmask = 4+64;
        break;
    case 2:
        dirmask = 1+4+16+64;
        break;
    case 3:
        if (gPartycleIndex & 1)
            dirmask = 2+32;
        else
            dirmask = 8+128;
        break;
    case 4:
        dirmask = 2+8+32+128;
        break;
    case 5:
        if (gPartycleIndex & 1)
            dirmask = 1+4+16+64;
        else
            dirmask = 2+8+32+128;
        break;
    case 6:
        dirmask = 1 << (gPartycleIndex & 7);
        break;
    case 7:
        dirmask = (1 << (gPartycleIndex & 7)) | (1 << ((gPartycleIndex >> 3) & 7));
        break;
    case 8:
        dirmask = physicsRand.genrand_int31() & 0xff;
        break;
    }
    return dirmask;
}


void spawn_partycle_circle(float x, float y, float r, int count, int mode)
{
    int j;
    float angle, inc;
    angle = 0;
    inc = (2 * M_PI) / count;
    for (j = 0; j < count; j++)
    {
        spawn_partycle(x + sin(angle) * r, y+ cos(angle) * r, gen_partycle_dirmask(mode));
        angle += inc;
    }
}

void spawn_partycle_grid(int xcount, int ycount, float dist, int mode)
{
    float xofs = 320 - (((xcount - 1) * dist) / 2);
    float yofs = 240 - (((ycount - 1) * dist) / 2);
    int i, j;

    for (i = 0; i < ycount; i++)
    {
        for (j = 0; j < xcount; j++)
        {
            spawn_partycle(xofs + j * dist, yofs + i * dist, gen_partycle_dirmask(mode));
        }
    }
}

void spawn_partycle_spiral(int count, int twists, int mode)
{
    int i;
    float angle = 0;
    float inc = (M_PI * 2 * twists) / count;
    float maxdist = 220.0f / count;
    for (i = 0; i < count; i++)
    {
        spawn_partycle(320 + sin(angle) * maxdist * (i + 4), 240 + cos(angle) * maxdist * (i + 4), gen_partycle_dirmask(mode));
        angle += inc;
    }

}


void spawn_partycle_random(float x, float y, int mode)
{
    spawn_partycle(x,y,gen_partycle_dirmask(mode));
}


void reset_stage(int stageno)
{
    int i;
    gPartycleIndex = 0;

#ifdef AUDIO_ENABLED
    if (mastermusicswitch)
    {
        FMUSIC_StopSong(s_midi[0]);
        FMUSIC_StopSong(s_midi[1]);
        FMUSIC_StopSong(s_midi[2]);
        FMUSIC_StopSong(s_midi[3]);
        FMUSIC_StopSong(s_midi[4]);
        FMUSIC_SetMasterVolume(s_midi[(stageno%4)+1], 0);
        FMUSIC_PlaySong(s_midi[(stageno % 4) + 1]);
    }
#endif



    for (i = 0; i < MAX_PARTYCLES; i++)
        partycle[i].live = 0;

    physicsRand.init_genrand(stageno);
    
    if (stageno < 10)
    {
        switch (stageno)
        {
        case 1:
            spawn_partycle(320-100,240-80,0);
            spawn_partycle(320+100,240-80,0);
            spawn_partycle(320,240-90,0);
            break;
        case 2:
            spawn_partycle(320-100,240-80,64);
            spawn_partycle(320+100,240-80,4);
            spawn_partycle(320,240-90,128);
            break;
        case 3:
            spawn_partycle_square(320,240,100,0);
            break;
        case 4:            
            spawn_partycle_square(320, 240, 50, 3);
            spawn_partycle_square(320, 240, 100, 2);
            spawn_partycle_square(320, 240, 150, 1);
            break;
        case 5:
            spawn_partycle_circle(320, 240, 180, 10, 4);
            break;
        case 6:
            for (i = 0; i < 4; i++)
                spawn_partycle_square(320 - 80 + physicsRand.genrand_int31()%160, 240 + physicsRand.genrand_int31()%120 - 60, physicsRand.genrand_int31()%100+50, 3);
            break;
        case 7:
            spawn_partycle_circle(320-120,240-120,40, 3, 3);
            spawn_partycle_circle(320+120,240-120,50, 4, 4);
            spawn_partycle_circle(320-120,240+120,60, 5, 5);
            spawn_partycle_circle(320+120,240+120,70, 6, 2);
            break;
        case 8:
            spawn_partycle_grid(4, 4, 100, 3);
            break;
        case 9:
            spawn_partycle_spiral(17, 2, 5);
            break;            
        }
    }
    else
    {
        switch (stageno % 5)
        {
        case 0:
            spawn_partycle_grid((stageno / 5) + 3, (stageno / 5) + 3, 100 - stageno, stageno % 8);
            break;
        case 1:
            spawn_partycle_spiral(stageno + 8, stageno % 3 + 1, stageno % 8);
            break;
        case 2:
            for (i = 0; i < stageno / 8; i++)
            {        
                spawn_partycle_circle(physicsRand.genrand_int31() % (640 - 128) + 64, physicsRand.genrand_int31() % (480 - 128) + 64, physicsRand.genrand_int31() % 50 + 100, stageno / 3 + physicsRand.genrand_int31() % (stageno / 6), stageno % 8);
            }
            break;
        case 3:
            for (i = 0; i < stageno / 4; i++)
            {        
                spawn_partycle_square(physicsRand.genrand_int31() % (640 - 128) + 64, physicsRand.genrand_int31() % (480 - 128) + 64, physicsRand.genrand_int31() % 50 + 100, stageno % 8);
            }
            break;
        case 4:
            for (i = 0; i < stageno / 5; i++)
                spawn_partycle_circle(320, 240, (i + 1) * 75 + 25, stageno / 2 + i * 5, (stageno + i) % 8);
            break;
        }

        int maxitout = stageno * 2;
        if (maxitout > MAX_PARTYCLES) maxitout = MAX_PARTYCLES;
        while (gPartycleIndex < maxitout)
        {        
            spawn_partycle_random(physicsRand.genrand_int31() % (640 - 32) + 16, physicsRand.genrand_int31() % (480 - 32) + 16, stageno % 8);
        }
        
    }

    for (i = 0; i < MAX_BULLETS; i++)
    {
        bullet[i].live = 0;
    }

    for (i = 0; i < MAX_POPUPS; i++)
    {
        if (popup[i].live)
        {
            free(popup[i].text);
            popup[i].text = 0;
            popup[i].live = 0;
        }
    }
    for (i = 0; i < MAX_PARTICLES; i++)
        particle[i].live = 0;

    gPlayerX = 320;
    gPlayerY = 240;
    gPlayerXi = 0;
    gPlayerYi = 0;
    gPlayerDirection = 0;
    gPlayerDirectionInc = 0;
    gStage = stageno;
    gKeyState = 0;
    last_tick = gStateStartTick = SDL_GetTicks();    

}



int mousehit(int x, int y, int w, int h)
{
    if (x < gMouseX &&
        x + w > gMouseX &&
        y < gMouseY &&
        y + h > gMouseY)
        return 1;
    return 0;
}


#define HANDLEKEY(x) if (down) gKeyState |= x; else gKeyState &= ~x;

void handle_key(int keysym, int down)
{
    switch(keysym)
    {
    case SDLK_ESCAPE:
        if (!down)
        {
            if (gGameState == STATE_MENU)
                exit(0);
            else 
            {
                gGameState = STATE_MENU;
                last_tick = gStateStartTick = SDL_GetTicks();
#ifdef AUDIO_ENABLED
                if (mastermusicswitch)
                {
                    FMUSIC_StopSong(s_midi[0]);
                    FMUSIC_StopSong(s_midi[1]);
                    FMUSIC_StopSong(s_midi[2]);
                    FMUSIC_StopSong(s_midi[3]);
                    FMUSIC_StopSong(s_midi[4]);
                    FMUSIC_PlaySong(s_midi[0]);
                }
#endif
            }
        }
        break;
    case SDLK_UP:
    case SDLK_w:
        HANDLEKEY(KEY_UP);
        break;
    case SDLK_DOWN:
    case SDLK_s:
        HANDLEKEY(KEY_DOWN);
        break;
    case SDLK_LEFT:
    case SDLK_a:
        HANDLEKEY(KEY_LEFT);
        break;
    case SDLK_RIGHT:
    case SDLK_d:
        HANDLEKEY(KEY_RIGHT);
        break;
    case SDLK_RCTRL:
    case SDLK_LCTRL:
        if (down) gKeyState |= KEY_FIRE;
        break;
    }
}

void process_events()
{
    SDL_Event event;

    while (SDL_PollEvent(&event)) 
    {
        switch (event.type) 
        {
        case SDL_KEYDOWN:
            handle_key(event.key.keysym.sym, 1);
            break;
        case SDL_KEYUP:
            handle_key(event.key.keysym.sym, 0);
            break;
        case SDL_MOUSEMOTION:
            gMouseX = (event.motion.x * 640) / screenwidth;
            gMouseY = (event.motion.y * 480) / screenheight;
            break;
        case SDL_MOUSEBUTTONUP:
            if (gGameState == STATE_MENU)
            {

                if (mousehit(50,150,8*8*2,20))
                {
                    gGameState = STATE_INGAME;
                    last_tick = gStateStartTick = SDL_GetTicks();
                    reset_stage(1);
                    gScore = 0;
#ifdef AUDIO_ENABLED
					FSOUND_PlaySound(FSOUND_FREE, s_click);
#endif

                }

                if (mousehit(50,200,12*8*2,20))
                {
                    gGameState = STATE_INGAME;
                    last_tick = gStateStartTick = SDL_GetTicks();
                    reset_stage(10);
                    gScore = 0;
#ifdef AUDIO_ENABLED
					FSOUND_PlaySound(FSOUND_FREE, s_click);
#endif
                }

                if (mousehit(50,250,8*8*2,20))
                {
                    gGameState = STATE_INGAME;
                    last_tick = gStateStartTick = SDL_GetTicks();
                    reset_stage(25);
                    gScore = 0;
#ifdef AUDIO_ENABLED
					FSOUND_PlaySound(FSOUND_FREE, s_click);
#endif
                }

                if (mousehit(50,300,5*8*2,20))
                {
                    gGameState = STATE_INGAME;
                    last_tick = gStateStartTick = SDL_GetTicks();
                    reset_stage(50);
                    gScore = 0;
#ifdef AUDIO_ENABLED
					FSOUND_PlaySound(FSOUND_FREE, s_click);
#endif
                }

                if (gStage > 0)
                {
                    if (mousehit(420,200,8*8*2,20))
                    {
                        gGameState = STATE_INGAME;
                        last_tick = gStateStartTick = SDL_GetTicks();
                        reset_stage(gStage);
                        gScore = 0;
#ifdef AUDIO_ENABLED
					    FSOUND_PlaySound(FSOUND_FREE, s_click);
#endif
                    }
                }

                if (mousehit(420,400,9*8*2,20))
                {
#ifdef AUDIO_ENABLED
					FSOUND_PlaySound(FSOUND_FREE, s_click);
#endif
                    mastermusicswitch = !mastermusicswitch;
                    if (mastermusicswitch)
                        FMUSIC_PlaySong(s_midi[0]);
                    else
                        FMUSIC_StopSong(s_midi[0]);
                }

                if (mousehit(50,400,9*8*2,20))
                {
                    exit(0);
                }
            }
            break;
        case SDL_QUIT:
            SDL_Quit();
            exit(0);
            break;
        }
    }
}

void drawrect(float x, float y, float w, float h, int color)
{
    glColor4f(((color >> 16) & 0xff) / 256.0f,
              ((color >> 8) & 0xff) / 256.0f,
              ((color >> 0) & 0xff) / 256.0f,
              ((color >> 24) & 0xff) / 256.0f);
    glBegin(GL_TRIANGLE_STRIP);
      glVertex2f(x,y);
      glVertex2f(x,y+h);
      glVertex2f(x+w,y);
      glVertex2f(x+w,y+h);
    glEnd();
}

void drawchar(int ch, float x, float y, float w, float h)
{
    // font = 6 rows, 16 chars per row
    float uofs = (ch & 15) * (1.0f / 16.0f) + 0.002f;
    float vofs = ((ch - 32) / 16) * (1.0f / 6.0f) + 0.002f;
    float uwid = (1.0f / 16.0f) - 0.004f;
    float vwid = (1.0f / 6.0f) - 0.004f;
    glBegin(GL_TRIANGLE_STRIP);
      glTexCoord2f(uofs,vofs);
      glVertex2f(x - w / 2,y - h / 2);
      glTexCoord2f(uofs,vofs+vwid);
      glVertex2f(x - w / 2,y + h / 2);
      glTexCoord2f(uofs+uwid,vofs);
      glVertex2f(x + w / 2,y - h / 2);
      glTexCoord2f(uofs+uwid,vofs+vwid);
      glVertex2f(x + w / 2,y + h / 2);
    glEnd();
}

void drawstring(char * string, float x, float y, int color, float alpha, float size = 1.0)
{
    glBindTexture(GL_TEXTURE_2D, tex_font);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glColor4f(((color >> 16) & 0xff) / 256.0f,
              ((color >> 8) & 0xff) / 256.0f,
              ((color >> 0) & 0xff) / 256.0f,
              alpha);
    while (*string)
    {
        drawchar(*string, x, y, 8 * size, 16 * size);
        string++;
        x += 8 * size;
    }
    glDisable(GL_TEXTURE_2D);
}

void drawsprite(int tex, float x, float y, float w, float h, float alpha = 0.95f)
{
    glBindTexture(GL_TEXTURE_2D, tex);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE);
    glColor4f(1.0f,1.0f,1.0f,alpha);
    glBegin(GL_TRIANGLE_STRIP);
      glTexCoord2f(0,0);
      glVertex2f(x - w / 2,y - h / 2);
      glTexCoord2f(0,1);
      glVertex2f(x - w / 2,y + h / 2);
      glTexCoord2f(1,0);
      glVertex2f(x + w / 2,y - h / 2);
      glTexCoord2f(1,1);
      glVertex2f(x + w / 2,y + h / 2);
    glEnd();
    glDisable(GL_TEXTURE_2D);
}

void drawplayer(float x, float y, float r)
{
    glBindTexture(GL_TEXTURE_2D, tex_ship);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glColor4f(1.0f,1.0f,1.0f,1.0f);
    float w = 32, h = 32;
    glPushMatrix();
    glTranslatef(x,y,0);
    glRotatef(r * 180.0f / M_PI ,0,0,1);
    glBegin(GL_TRIANGLE_STRIP);
      glTexCoord2f(0,0);
      glVertex2f(- w / 2,- h / 2);
      glTexCoord2f(0,1);
      glVertex2f(- w / 2, h / 2);
      glTexCoord2f(1,0);
      glVertex2f( w / 2,- h / 2);
      glTexCoord2f(1,1);
      glVertex2f( w / 2, h / 2);
    glEnd();
    glPopMatrix();
    glDisable(GL_TEXTURE_2D);
}

void drawpartycle(float x, float y, float w, float h, int directions, int tick)
{
    glBindTexture(GL_TEXTURE_2D, tex_base);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE);
    glColor4f(1.0f,1.0f,1.0f,.95f);
    glBegin(GL_TRIANGLE_STRIP);
      glTexCoord2f(0,0);
      glVertex2f(x - w / 2,y - h / 2);
      glTexCoord2f(0,1);
      glVertex2f(x - w / 2,y + h / 2);
      glTexCoord2f(1,0);
      glVertex2f(x + w / 2,y - h / 2);
      glTexCoord2f(1,1);
      glVertex2f(x + w / 2,y + h / 2);
    glEnd();
    int i; 
    for (i = 0; i < 8; i++)
    {
        if (directions & (1 << i))
        {
            glColor4f(1.0f,1.0f,1.0f,sin((tick + i * 20 + w + h) * 0.003134) * 0.3f + 0.4f);
            glBindTexture(GL_TEXTURE_2D, tex_dir[i]);
            glBegin(GL_TRIANGLE_STRIP);
            glTexCoord2f(0,0);
            glVertex2f(x - w / 2,y - h / 2);
            glTexCoord2f(0,1);
            glVertex2f(x - w / 2,y + h / 2);
            glTexCoord2f(1,0);
            glVertex2f(x + w / 2,y - h / 2);
            glTexCoord2f(1,1);
            glVertex2f(x + w / 2,y + h / 2);
            glEnd();
        }
    }

    glDisable(GL_TEXTURE_2D);
}

void spawn_bullet(float x, float y, float dir, int value, int tex = tex_red)
{
    int i = 0;
    while (i < MAX_BULLETS && bullet[i].live) i++;
    if (i >= MAX_BULLETS) return;
    bullet[i].live = 1;
    bullet[i].x = x;
    bullet[i].y = y;
    bullet[i].dirx = sin(dir) * 2 * ((value == 1) ? 2 : 1);
    bullet[i].diry = -cos(dir) * 2  * ((value == 1) ? 2 : 1);
    bullet[i].value = value;
    bullet[i].tex = tex;
}

void spawn_particle(float x, float y, int tex = tex_base)
{
    int i = visualRand.genrand_int31() % MAX_PARTICLES;
    if (particle[i].live)
        return;
    particle[i].live = 100 + visualRand.genrand_int31() % 100;
    particle[i].x = x;
    particle[i].y = y;
    float r = (visualRand.genrand_int31() % 1000) / 1000.0f;
    float d = (visualRand.genrand_int31() % 1000) / 1000.0f;
    particle[i].dirx = sin(r * 2 * M_PI) * 0.25 * (1 + d);
    particle[i].diry = -cos(r * 2 * M_PI) * 0.25 * (1 + d);    
    particle[i].tex = tex;
}

void spawn_bullets(int partycleno, int value)
{
    if (value > 1024) value = 1024;
    int i;
    for (i = 0; i < 8; i++)
    {
        if (value == 0 || partycle[partycleno].dirmask & (1 << i))
        {   
            if (value == 0)
                spawn_bullet(partycle[partycleno].x, partycle[partycleno].y, i * M_PI / 4, value, tex_black);
            else
                spawn_bullet(partycle[partycleno].x, partycle[partycleno].y, i * M_PI / 4, value);
        }
    }
}

static void draw_screen_ingame()
{
    int tick = SDL_GetTicks();

    int i, j, k;    

    if (tick - last_tick > 1000) last_tick = tick;

    while (last_tick < tick)
    {
        for (i = 0; i < MAX_PARTICLES; i++)
        {
            if (particle[i].live)
            {
                particle[i].live--;
                particle[i].x += particle[i].dirx;
                particle[i].y += particle[i].diry;
            }
        }        

        if (gGameState == STATE_INGAME || gGameState == STATE_STAGECLEAR)
        {
            gPlayerDirection += gPlayerDirectionInc;
            gPlayerX += gPlayerXi;
            gPlayerY += gPlayerYi;
            if (gKeyState & KEY_UP) 
            {
                gPlayerXi += sin(gPlayerDirection) * 0.071;
                gPlayerYi += -cos(gPlayerDirection) * 0.071;
                spawn_particle(gPlayerX, gPlayerY, tex_red);
            }
            if (gKeyState & KEY_DOWN)
            {
                gPlayerXi -= sin(gPlayerDirection) * 0.051;
                gPlayerYi -= -cos(gPlayerDirection) * 0.051;
                spawn_particle(gPlayerX, gPlayerY, tex_red);
            }
            if (gKeyState & KEY_LEFT)
            {
                gPlayerDirectionInc -= 0.005;
            }
            if (gKeyState & KEY_RIGHT)
            {
                gPlayerDirectionInc += 0.005;
            }
            
            gPlayerDirectionInc *= 0.9;
            gPlayerXi *= 0.965;
            gPlayerYi *= 0.965;
            if (gKeyState & KEY_FIRE)
            {
                gKeyState &= ~KEY_FIRE;
                spawn_bullet(gPlayerX + sin(gPlayerDirection) * 16, gPlayerY - cos(gPlayerDirection) * 16, gPlayerDirection, 1, tex_green);
#ifdef AUDIO_ENABLED
				FSOUND_PlaySound(FSOUND_FREE, s_shoot);
#endif

            }
            
            int collided = 0;

            for (i = 0; i < MAX_PARTYCLES; i++)
            {            
                if (partycle[i].live)
                {
                    float xd = partycle[i].x-gPlayerX;
                    float yd = partycle[i].y-gPlayerY;
                    float d = sqrt(xd*xd+yd*yd);
                    if (d < 20)
                    {
                        xd /= d;
                        yd /= d;
                        gPlayerX = partycle[i].x - xd * 20;
                        gPlayerY = partycle[i].y - yd * 20;
                        collided = 1;
                    }
                }
            }

            if (gPlayerX < 16) 
            {
                gPlayerX = 16;
                collided = 1;
            }
            if (gPlayerX > 640-16)  
            {
                gPlayerX = 640-16;
                collided = 1;
            }
            if (gPlayerY < 16)  
            {
                gPlayerY = 16;
                collided = 1;
            }
            if (gPlayerY > 480-16)  
            {
                gPlayerY = 480-16;
                collided = 1;
            }
        }


        for (i = 0; i < MAX_BULLETS; i++)
        {
            if (bullet[i].live)
            {
                bullet[i].x += bullet[i].dirx;
                bullet[i].y += bullet[i].diry;

                if (gGameState == STATE_INGAME || gGameState == STATE_STAGECLEAR)
                {
                    if (bullet[i].value != 1 && sqrt((bullet[i].x-gPlayerX) * (bullet[i].x-gPlayerX) +
                                (bullet[i].y-gPlayerY) * (bullet[i].y-gPlayerY)) < 8)
                    {
#ifdef AUDIO_ENABLED
    					FSOUND_PlaySound(FSOUND_FREE, s_gameover);
#endif
                        gGameState = STATE_DYING; // player dies
                        last_tick = gStateStartTick = tick;
                        for (k = 0; k < PARTICLES_PER_EXPLOSION * 4; k++)
                            spawn_particle(gPlayerX, gPlayerY, tex_red);
                    }
                }


                for (j = 0; j < MAX_PARTYCLES; j++)
                {
                    if (partycle[j].live)
                    {
                        if (sqrt((bullet[i].x-partycle[j].x) * (bullet[i].x-partycle[j].x) +
                                 (bullet[i].y-partycle[j].y) * (bullet[i].y-partycle[j].y)) < 16)
                        {
                            if (gGameState == STATE_INGAME)
                            {
#ifdef AUDIO_ENABLED
                                if (bullet[i].value)
                                {
                                    int freq = 11000 + bullet[i].value * 1000;
                                    if (freq > 88000) freq = 88000;

            					    int c = FSOUND_PlaySound(FSOUND_FREE, s_pop);
                                    FSOUND_SetFrequency(c, freq);
                                }
                                else
                                    FSOUND_PlaySound(FSOUND_FREE, s_badpop);
#endif
                                
                                gScore += bullet[i].value * gStage;
                                char temp[200];
                                sprintf(temp,"%dx", bullet[i].value);
                                spawn_popup(temp, partycle[j].x, partycle[j].y);
                            }
                            partycle[j].live = 0;
                            bullet[i].live = 0;
                            spawn_bullets(j, bullet[i].value * 2);
                            for (k = 0; k < PARTICLES_PER_EXPLOSION; k++)
                                spawn_particle(partycle[j].x, partycle[j].y);                            
                        }
                    }
                }

                if (bullet[i].x < 0 ||
                    bullet[i].y < 0 ||
                    bullet[i].x > 640 ||
                    bullet[i].y > 480)
                    bullet[i].live = 0;
            }
        }
        
        int livepartycles = 0;
        for (j = 0; j < MAX_PARTYCLES; j++)
        {
            if (partycle[j].live)
            {
                partycle[j].live--;
                if (partycle[j].live == 0)
                {
                    spawn_bullets(j, 0);
                    for (k = 0; k < PARTICLES_PER_EXPLOSION; k++)
                        spawn_particle(partycle[j].x, partycle[j].y);                            
#ifdef AUDIO_ENABLED
                    FSOUND_PlaySound(FSOUND_FREE, s_badpop);
#endif
                }
                else
                {
                    livepartycles++;
                }
            }        
        }

        if (livepartycles == 0)
        {
            if (gGameState == STATE_INGAME)
            {
                gGameState = STATE_STAGECLEAR;
                last_tick = gStateStartTick = tick;
            }
        }

        for (j = 0; j < MAX_POPUPS; j++)
        {
            if (popup[j].live)
            {
                popup[j].live--;
                popup[j].y -= 0.05f;
                if (!popup[j].live)
                {
                    free(popup[j].text);
                    popup[j].text = 0;
                }
            }
        }


        last_tick += 1000 / 100;
    }

/////////////////////////////////////////////////////////////////////////////////////////

    float sizer4k = ((tick - gStateStartTick) / 4000.0f);
    float sizer8k = ((tick - gStateStartTick) / 8000.0f);
    float sizer2k = ((tick - gStateStartTick) / 1000.0f);
    if (sizer8k > 1.0) sizer8k = 1.0;
    if (sizer4k > 1.0) sizer4k = 1.0;
    if (sizer2k > 1.0) sizer2k = 1.0;
    sizer2k = 1.0 - sizer2k;
    sizer2k = sizer2k * sizer2k * sizer2k;
    sizer2k = 1.0 - sizer2k;

    if (gGameState != STATE_INGAME)
    {
        sizer8k = sizer2k = sizer4k = 1.0;
    }

    glClearColor(0,0,0,1.0);
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    drawsprite(tex_bg,320 + sin(tick * 0.00012134) * 10,240 + sin(tick * 0.00012234) * 10,640+20,480+20,1.0);

    for (i = 0; i < MAX_PARTICLES; i++)
    {
        if (particle[i].live)
            drawsprite(particle[i].tex, particle[i].x, particle[i].y, particle[i].live * (7 / 100.0f), particle[i].live * (7 / 100.0f), 0.5f);
    }

    for (i = 0; i < MAX_PARTYCLES; i++)
    {
        if (partycle[i].live)
        {
            if (partycle[i].live < 500)
            {
                float shakey = (500 - partycle[i].live) / 500.0f;
                shakey = sqrt(shakey);
                drawpartycle(partycle[i].x + ((visualRand.genrand_int31() % 100) / 20.0f) * shakey, partycle[i].y + ((visualRand.genrand_int31() % 100) / 20.0f) * shakey, 32, 32, partycle[i].dirmask, tick);
            }
            else
            {
                drawpartycle(partycle[i].x, partycle[i].y, 32 * sizer2k, 32 * sizer2k, partycle[i].dirmask, tick);
            }
        }
    }

    for (i = 0; i < MAX_BULLETS; i++)
    {
        if (bullet[i].live)
        {
            float size = 7 + sin(tick * 0.053);
            drawsprite(bullet[i].tex, bullet[i].x - bullet[i].dirx*7, bullet[i].y - bullet[i].diry*7, size-1, size-1, 0.25f);
            drawsprite(bullet[i].tex, bullet[i].x - bullet[i].dirx*3, bullet[i].y - bullet[i].diry*3, size, size, 0.5f);
            drawsprite(bullet[i].tex, bullet[i].x, bullet[i].y, size, size);
        }
    }

    if (gGameState == STATE_INGAME || gGameState == STATE_STAGECLEAR)
        drawplayer(gPlayerX,gPlayerY,gPlayerDirection + (1.0 - sizer2k) * 20);

    for (j = 0; j < MAX_POPUPS; j++)
    {
        if (popup[j].live)
        {
            drawstring(popup[j].text,popup[j].x+2,popup[j].y+2,0,0.5f);
            drawstring(popup[j].text,popup[j].x,popup[j].y,0xffff7f,0.75f);
        }
    }

    char temp[200];
    sprintf(temp,"%08d", gScore);
    drawstring(temp, 20,20,0,0.5f);
    drawstring(temp, 20-2,20-2,0xffffff,0.75f);

    sprintf(temp,"Stage %d", gStage);
    drawstring(temp, 556,20,0,0.5f);
    drawstring(temp, 556-2,20-2,0xffffff,0.75f);

    if (gGameState == STATE_INGAME)
    {
        if (tick - gStateStartTick < 8000)
        {
            char *hint = NULL;

            switch (gStage)
            {
            case 1:
                hint = "Move with cursors, fire with Ctrl (WASD also works)";
                break;
            case 2:
                hint = "Partycles with arrows spit out bullets";
                break;
            case 3:
                hint = "Chain reactions double your points - 2x, 4x, 8x, 16x..";
                break;
            case 4:
                hint = "You can also move in reverse";
                break;
            case 5:
                hint = "Partycles pop by themselves too, but you don't get points";
                break;
            case 10:
                hint = "Intermediate!";
                break;
            case 15:
                hint = "For good time, call http://iki.fi/sol/";
                break;
            case 25:
                hint = "Advanced!";
                break;
            case 30:
                hint = "Learn to make games like these at http://iki.fi/sol/";
                break;
            case 50:
                hint = "Crazy!";
                break;
            }


            if (hint)
            {
                int len = strlen(hint) / 2;
                drawstring(hint, 320-len*(8 * (sizer8k * 0.25 + 1)),350 + sizer8k * 10,0,0.75 - (sizer8k * 0.75),1 + sizer8k*0.25);
                drawstring(hint, 320-len*(8 * (sizer8k * 0.25 + 1)),350 + sizer8k * 10,0xFFFFFF,1.0 - sizer8k,1 + sizer8k*0.25);
            }

        }

        if (tick - gStateStartTick < 4000)
        {
            char temp[200];        
            sprintf(temp, "Stage %d", gStage);
            drawstring(temp, 320-(4*8*(sizer4k + 1.6)),240-8,0,0.75 - (sizer2k * 0.75),sizer4k + 1.6);
            drawstring(temp, 320-(4*8*(sizer4k + 1.5)),240-8,0xffddaa,1.0 - sizer4k,sizer4k + 1.5);

            sizer4k = 1.0 - sizer4k;
            drawrect(0,0,640,480,(int)((sizer4k * sizer4k * sizer4k * sizer4k) * 0xff) << 24);
#ifdef AUDIO_ENABLED
            if (mastermusicswitch)
                FMUSIC_SetMasterVolume(s_midi[(gStage%4)+1], 32 * (1 - sizer4k));
#endif
        }
    }



    if (gGameState == STATE_STAGECLEAR)
    {
        float sizer = ((tick - gStateStartTick) / 4000.0f);
        if (sizer > 1.0f) sizer = 1.0f;
        drawstring("Stage Clear", 320-(5*8*(-sizer + 2.1)),240-8,0,0.75,-sizer + 2.1);
        drawstring("Stage Clear", 320-(5*8*(-sizer + 2.0)),240-8,0xffddaa,1.0 - sizer,-sizer + 2.0);
        if (tick - gStateStartTick > 4000)
        {
            gGameState = STATE_INGAME;
            reset_stage(gStage+1);
        }
        drawrect(0,0,640,480,(int)((sizer * sizer * sizer * sizer) * 0xff) << 24);
#ifdef AUDIO_ENABLED
        if (mastermusicswitch)
            FMUSIC_SetMasterVolume(s_midi[(gStage%4)+1], 32 * (1 - sizer));
#endif
    }

    if (gGameState == STATE_DYING)
    {
        float sizer = ((tick - gStateStartTick) / 4000.0f);
        if (sizer > 1.0f) sizer = 1.0f;
        drawstring("Game over", 320-(4*8*(sizer * 2 + 2.1)),240-8,0,0.75,sizer * 2 + 2.1);
        drawstring("Game over", 320-(4*8*(sizer * 2 + 2.0)),240-8,0xff0000,sizer,sizer * 2 + 2.0);
        drawrect(0,0,640,480,(int)((sizer * sizer * sizer * sizer) * 0xff) << 24);
#ifdef AUDIO_ENABLED
        if (mastermusicswitch)
            FMUSIC_SetMasterVolume(s_midi[(gStage%4)+1], 32 * (1 - sizer));
#endif
        if (tick - gStateStartTick > 4000)
        {
            gGameState = STATE_MENU;
            last_tick = gStateStartTick = tick;
#ifdef AUDIO_ENABLED
            if (mastermusicswitch)
            {
                FMUSIC_StopSong(s_midi[0]);
                FMUSIC_StopSong(s_midi[1]);
                FMUSIC_StopSong(s_midi[2]);
                FMUSIC_StopSong(s_midi[3]);
                FMUSIC_StopSong(s_midi[4]);
                FMUSIC_PlaySong(s_midi[0]);
            }
#endif

        }
    }

    SDL_Delay(1);
	SDL_GL_SwapBuffers();
}


void draw_screen_menu()
{
    int tick = SDL_GetTicks();
    float fadein = (tick - gStateStartTick) / 1000.0f;
    if (fadein > 1.0) fadein = 1.0;
    glClearColor(0,0,0,1.0);
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    drawsprite(tex_bg,320,240,640,480,fadein);

    drawrect(0, 76, 640 * fadein, 2, 0xffffddaa);

    drawstring(TITLE, 43, 63*fadein, 0, 0.75, 3.0);
    drawstring(TITLE, 40, 60*fadein, 0xffddaa, 1.0, 3.0);
    drawstring("http://iki.fi/sol/",490 + 150 * (1.0 - fadein),88,0xffddaa,1.0,1.0);

    if (gScore != 0)
    {
        char temp[200];
        sprintf(temp,"Latest score: %08d", gScore);
        drawstring(temp,450 + 400 - 400*fadein,10,0xffddaa,1.0);
        sprintf(temp,"from stage %d", gStage);
        drawstring(temp,422 + 400 - 400 * fadein, 222,0,0.75);
        drawstring(temp,420 + 400 - 400 * fadein, 220,0xaaddff,1.0);
    }

    drawstring("Beginner", 53 - 100 + 100 * fadein, 153, 0,0.5,2);
    if (mousehit(50,150,8*8*2,20))
        drawstring("Beginner", 50 - 100 + 100 * fadein, 150, 0xffffff,1,2);
    else
        drawstring("Beginner", 50 - 100 + 100 * fadein, 150, 0xaaddff,1,2);

    drawstring("Intermediate", 53 - 200 + 200 * fadein, 203, 0,0.5,2);
    if (mousehit(50,200,12*8*2,20))
        drawstring("Intermediate", 50 - 200 + 200 * fadein, 200, 0xffffff,1,2);
    else
        drawstring("Intermediate", 50 - 200 + 200 * fadein, 200, 0xaaddff,1,2);

    drawstring("Advanced", 53 - 300 + 300 * fadein, 253, 0,0.5,2);
    if (mousehit(50,250,8*8*2,20))
        drawstring("Advanced", 50 - 300 + 300 * fadein, 250, 0xffffff,1,2);
    else
        drawstring("Advanced", 50 - 300 + 300 * fadein, 250, 0xaaddff,1,2);

    drawstring("Crazy", 53 - 400 + 400 * fadein, 303, 0,0.5,2);
    if (mousehit(50,300,5*8*2,20))
        drawstring("Crazy", 50 - 400 + 400 * fadein, 300, 0xffffff,1,2);
    else
        drawstring("Crazy", 50 - 400 + 400 * fadein, 300, 0xaaddff,1,2);

    if (gStage > 1)
    {
        drawstring("Continue", 423 + 400 - 400 * fadein, 203, 0,0.5,2);
        if (mousehit(420,200,8*8*2,20))
            drawstring("Continue", 420 + 400 - 400 * fadein, 200, 0xffffff,1,2);
        else
            drawstring("Continue", 420 + 400 - 400 * fadein, 200, 0xaaddff,1,2);
    }

    drawstring(mastermusicswitch?"Music on":"Music off", 423 + 500 - 500 * fadein, 403, 0,0.5,2);
    if (mousehit(420,400,9*8*2,20))
        drawstring(mastermusicswitch?"Music on":"Music off", 420 + 500 - 500 * fadein, 400, 0xffffff,1,2);
    else
        drawstring(mastermusicswitch?"Music on":"Music off", 420 + 500 - 500 * fadein, 400, 0xaaddff,1,2);


    drawstring("Exit game", 53 - 500 + 500 * fadein, 403, 0,0.5,2);    
    if (mousehit(50,400,9*8*2,20))
        drawstring("Exit game", 50 - 500 + 500 * fadein, 400, 0xffffff,1,2);
    else
        drawstring("Exit game", 50 - 500 + 500 * fadein, 400, 0xaaddff,1,2);

    drawplayer(gMouseX, gMouseY, M_PI * 1.75);

    if (fadein < 1.0)
    {
        float sizer = 1.0 - fadein;
        drawrect(0,0,640,480,(int)((sizer * sizer * sizer * sizer) * 0xff) << 24);
    }

    SDL_Delay(1);
	SDL_GL_SwapBuffers();
}

void draw_screen()
{
    if (gGameState == STATE_MENU)
        draw_screen_menu();
    else
        draw_screen_ingame();
}

int main (int argc, char** args)
{
    const SDL_VideoInfo *info = NULL;
    int bpp = 0;
    int flags = 0;
    int i;

    visualRand.init_genrand(0xc0cac01a);

#ifdef AUDIO_ENABLED
	// Init fmod
	if (FSOUND_GetVersion() < FMOD_VERSION)
		exit(0);
	if (!FSOUND_Init(32000, 64, 0))
		exit(0);
	
	// Load sounds
    s_badpop = FSOUND_Sample_Load(FSOUND_UNMANAGED, "badpop_new.wav", FSOUND_NORMAL, 0, 0);
    s_click = FSOUND_Sample_Load(FSOUND_UNMANAGED, "click_new.wav", FSOUND_NORMAL, 0, 0);
    s_gameover = FSOUND_Sample_Load(FSOUND_UNMANAGED, "gameover_new.wav", FSOUND_NORMAL, 0, 0);
    s_pop = FSOUND_Sample_Load(FSOUND_UNMANAGED, "pop_new2.wav", FSOUND_NORMAL, 0, 0);
    s_shoot = FSOUND_Sample_Load(FSOUND_UNMANAGED, "shoot_new.wav", FSOUND_NORMAL, 0, 0);
    s_midi[0] = FMUSIC_LoadSong("menu.mid");
    s_midi[1] = FMUSIC_LoadSong("partycle1.mid");
    s_midi[2] = FMUSIC_LoadSong("partycle2.mid");
    s_midi[3] = FMUSIC_LoadSong("partycle3.mid");
    s_midi[4] = FMUSIC_LoadSong("partycle4.mid");
    for (i = 0; i < 5; i++)
    {
        FMUSIC_SetMasterVolume(s_midi[i], 32);
        FMUSIC_SetLooping(s_midi[i], 1);
//        FMUSIC_PlaySong(s_midi);
    }
    FMUSIC_PlaySong(s_midi[0]);


#endif

    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE) < 0) 
    {
        fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
        SDL_Quit();
        exit(0);
    }

    info = SDL_GetVideoInfo();

    if (!info) 
    {
        fprintf(stderr, "Video query failed: %s\n", SDL_GetError());
        SDL_Quit();
        exit(0);
    }

#ifdef _DEBUG
    int fsflag = 0;
#else
    int fsflag = 1;
#endif

    if (argc > 1) fsflag = !fsflag;

    if (fsflag) 
    {
        screenwidth = info->current_w;
        screenheight = info->current_h;
        bpp = info->vfmt->BitsPerPixel;
        flags = SDL_OPENGL | SDL_FULLSCREEN;
    }
    else
    {
        screenwidth = 800;
        screenheight = 600;
        bpp = info->vfmt->BitsPerPixel;
        flags = SDL_OPENGL;
    }

    SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

    if (SDL_SetVideoMode(screenwidth, screenheight, bpp, flags) == 0) 
    {
        fprintf( stderr, "Video mode set failed: %s\n", SDL_GetError());
        SDL_Quit();
        exit(0);
    }

    SDL_WM_SetCaption(TITLE " - http://iki.fi/sol/", NULL);
    
    float aspect = 4.0f/3.0f;
    if (screenheight / screenwidth > 4.0f / 3.0f)
    {
        float realx = screenheight * aspect;
        float extrax = screenwidth - realx;

        glViewport( extrax / 2, 0, realx, screenheight );
    }
    else
    {
        float realy = screenwidth / aspect;
        float extray = screenheight - realy;

        glViewport( 0, extray / 2, screenwidth, realy );
    }

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );
	gluOrtho2D(0,640,480,0);

    tex_bg = load_texture("bgmess.png");
    tex_ship = load_texture("shippy.png");
    tex_font = load_texture("fona.png");
    tex_red = load_texture("red.png");
    tex_black = load_texture("black.png");
    tex_green = load_texture("green.png");
    tex_base = load_texture("base.png");
    tex_dir[0] = load_texture("n.png");
    tex_dir[1] = load_texture("ne.png");
    tex_dir[2] = load_texture("e.png");
    tex_dir[3] = load_texture("se.png");
    tex_dir[4] = load_texture("s.png");
    tex_dir[5] = load_texture("sw.png");
    tex_dir[6] = load_texture("w.png");
    tex_dir[7] = load_texture("nw.png");

    SDL_ShowCursor(0);
    
    while (1) 
    {
        process_events();
        draw_screen();
    }

    return 0;
}
